/* Copyright (C) 2024 Facility for Antiproton and Ion Research in Europe, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Dominik Smith [committer], Alexandru Bercuci */

#include "HitFinder2D.h"

#include "log.hpp"

#include <numeric>

using std::cout;
using std::endl;
using std::vector;

namespace cbm::algo::trd
{

  //////////// TO DO: fdEdep is not used!!! Will be needed at some point.
  double HitFinder2D::fgDT[] = {4.181e-6, 1586, 24};

  //_______________________________________________________________________________
  HitFinder2D::HitFinder2D(HitFinder2DModPar params) : fParams(params), fHit(params.rowPar[0].padPar.size()) {}

  //_______________________________________________________________________________
  std::vector<Hit> HitFinder2D::operator()(std::vector<Cluster2D>* clusters)
  {
    const int nclusters = clusters->size();
    std::vector<Hit> hits;

    for (int icluster = 0; icluster < nclusters; icluster++) {
      Cluster2D* cluster = &clusters->at(icluster);
      auto hit           = MakeHit(icluster, cluster, &cluster->GetDigis(), nclusters);
      if (hit.Address() == -1) continue;
      hits.push_back(hit);
    }

    int a0, nhits = hits.size();
    float Dx(2 * fParams.padSizeX), Dy(2 * fParams.padSizeY);
    for (int ih(0); ih < nhits; ih++) {
      Hit* h0 = &hits[ih];
      if (h0->IsUsed()) continue;

      for (int jh(ih + 1); jh < nhits; jh++) {
        Hit* h1 = &hits[jh];
        if (h1->IsUsed()) continue;

        // basic check on Time
        if (h1->Time() < 4000 - h0->Time()) continue;  // TODO check with preoper time calibration
        // skip next hits for being too far (10us) in the future
        if (h1->Time() > 10000 + h0->Time()) break;

        // basic check on Row
        if (std::abs(h1->Y() - h0->Y()) > Dy) continue;

        // basic check on Col
        if (std::abs(h1->X() - h0->X()) > Dx) continue;

        // go to topologic checks
        if (!(a0 = CheckMerge(h0->GetRefId(), h1->GetRefId()))) continue;

        ProjectDigis(a0 < 0 ? h0->GetRefId() : h1->GetRefId(), a0 < 0 ? h1->GetRefId() : h0->GetRefId());

        // call the working algorithm
        if (MergeHits(h0, a0)) {
          h0->SetRowCross((bool) h1);
          h1->SetRefId(-1);
        }
      }
    }

    const auto ret = std::remove_if(hits.begin(), hits.end(), [](const auto& obj) { return obj.IsUsed(); });
    hits.erase(ret, hits.end());

    // clear all calibrated digis
    for (auto& vec : fDigis) {
      for (auto& digiPtr : vec)
        delete digiPtr;
      vec.clear();
    }
    fDigis.clear();

    return hits;
  }

  //_______________________________________________________________________________
  int HitFinder2D::CheckMerge(int cid, int cjd)
  {
    /** Check topologic conditions if the 2 clusters (in digi representation) can be merged.
 * The first pair is always from the bottom row
 * return the anode candidate wrt boundary 1, 2, 3 for the first 3 anodes of the upper row; -1, -2, -3 for the bottom row (r0) or 0 if the check fails
 */

    bool on(kFALSE);
    int vc[2] = {-1, -1}, vm[2] = {0};
    double M[2] = {-1., -1.}, S[2] = {0.};
    vector<DigiRec*>::iterator jp[2];

    for (int rowId(0); rowId < 2; rowId++) {
      double rtMax      = 0.;
      double mdMax      = 0.;
      const int vcid[2] = {cid, cjd};
      for (auto id = fDigis[vcid[rowId]].begin(); id != fDigis[vcid[rowId]].end(); id++) {
        int col;
        GetPadRowCol((*id)->GetAddressChannel(), col);

        // mark max position and type
        const double t = (*id)->GetTiltCharge(on);
        if (on && t > rtMax) {
          vc[rowId] = col;
          vm[rowId] = 0;
          rtMax     = t;
        }
        const double r = (*id)->GetRectCharge(on);
        if (on && r > rtMax) {
          vc[rowId] = col;
          vm[rowId] = 1;
          rtMax     = r;
        }

        double m = 0.;
        double d = 0.;
        if (!rowId) {  // compute TR pairs on the bottom row
          m = 0.5 * (t + r);
          d = r - t;
        }
        else {  // compute RT pairs on the upper row
          auto jd  = std::next(id);
          double T = 0;
          if (jd != fDigis[vcid[rowId]].end()) T = (*jd)->GetTiltCharge(on);
          m = 0.5 * (r + T);
          d = r - T;
        }
        if (std::abs(m) > 0.) d = 1.e2 * d / m;
        if (m > mdMax) {
          mdMax     = m;
          M[rowId]  = m;
          S[rowId]  = d;
          jp[rowId] = id;
        }
      }
    }
    int rowMax = M[0] > M[1] ? 0 : 1;

    // basic check on col of the max signal
    const int dc = vc[1] - vc[0];
    if (dc < 0 || dc > 1) return 0;

    // special care for both tilt maxima :  the TT case
    // recalculate values on the min row on neighbor column
    if (!vm[0] && !vm[1]) {
      if (rowMax == 0) {  // modify r=1
        double r = 0.;
        double T = 0.;
        if (M[1] >= 0) {
          if (jp[1] != fDigis[cjd].end()) jp[1]++;
          if (jp[1] != fDigis[cjd].end()) {
            r = (*jp[1])->GetRectCharge(on);
            jp[1]++;
            if (jp[1] != fDigis[cjd].end()) T = (*jp[1])->GetTiltCharge(on);
          }
        }
        M[1] = 0.5 * (r + T);
        S[1] = r - T;
      }
      else {  // modify r=0
        double r = 0.;
        double t = 0.;
        if (M[0] >= 0) {
          if (jp[0] != fDigis[cid].begin()) jp[0]--;
          if (jp[0] != fDigis[cid].begin()) {
            r = (*jp[0])->GetRectCharge(on);
            t = (*jp[0])->GetTiltCharge(on);
          }
        }
        M[0] = 0.5 * (t + r);
        S[0] = r - t;
      }
    }
    rowMax = M[0] > M[1] ? 0 : 1;

    // Build the ratio of the charge
    const float mM = M[rowMax ? 0 : 1] / M[rowMax];
    const float mS = std::abs(S[rowMax]), mM_l[3] = {0.15, 0.5, 1}, mM_r[3] = {0, 0.28, 1}, mS_r[3] = {43, 27, 20};
    float dSdM[2], S0[2];

    for (int i(0); i < 2; i++) {
      dSdM[i] = (mS_r[i + 1] - mS_r[i]) / (mM_r[i + 1] - mM_r[i]);
      S0[i]   = mS_r[i] - dSdM[i] * mM_r[i];
    }
    const int irange = mM < mM_r[1] ? 0 : 1;
    if (mS > S0[irange] + dSdM[irange] * mM) return 0;

    for (int ia(0); ia < 3; ia++) {
      if (mM < mM_l[ia]) return (rowMax ? 1 : -1) * (3 - ia);
    }
    return 0;
  }


  //_______________________________________________________________________________
  bool HitFinder2D::MergeHits(Hit* h, int a0)
  {
    int n0(fHit.fSignal.size() - 2);
    auto [dx, dy] = fHit.GetDxDy(n0);

    int typ(fHit.GetHitClass());
    // get position correction [pw]
    double xcorr = fHit.GetXcorr(dx, typ, 1) / fParams.padSizeX, xcorrBias(xcorr);
    if (fHit.IsBiasX()) {
      typ      = fHit.GetHitRcClass(a0);
      int xmap = fHit.vyM & 0xff;
      switch (n0) {
        case 4:
          if (dx < 0)
            xcorrBias += (fHit.IsBiasXleft() ? -0.12 : 0.176);
          else
            xcorrBias += (xmap == 53 || xmap == 80 || xmap == 113 || xmap == 117 ? -0.176 : 0.12);
          break;
        case 5:
        case 7:
          if (typ < 0) break;
          if (xmap == 50 || xmap == 58 || xmap == 146 || xmap == 154) {
            if (typ == 2)
              xcorr += 0.0813;
            else if (typ == 3) {
              xcorr -= 0.0813;
              typ = 2;
            }
            dx -= xcorr;
            fHit.RecenterXoffset(dx);
            xcorrBias = fHit.GetXcorr(dx, typ, 2) / fParams.padSizeX;
          }
          else {
            dx -= xcorr;
            fHit.RecenterXoffset(dx);
            if (typ < 2)
              xcorrBias = fHit.GetXcorr(dx, typ, 2) / fParams.padSizeX;
            else if (typ == 2)
              xcorrBias = 0.12;
            else if (typ == 3)
              xcorrBias = -0.12;
          }
          break;
        default:
          if (typ < 0)
            break;
          else if (typ == 2)
            xcorr += 0.0813;
          else if (typ == 3) {
            xcorr -= 0.0813;
            typ = 2;
          }
          dx -= xcorr;
          fHit.RecenterXoffset(dx);
          xcorrBias = fHit.GetXcorr(dx, typ, 2) / fParams.padSizeX;
          break;
      }
    }
    else {
      if (typ) xcorrBias += (dx < 0 ? 1 : -1) * 0.0293;
    }
    std::tie(dx, dy) = CorrectPosition(dx, dy, xcorrBias);

    double edx(1), edy(1), edt(60), time(-21), tdrift(100), e(200);
    CalibrateHit(h, dx, dy, edx, edy, edt, time, tdrift, e);

    return kTRUE;
  }

  //_______________________________________________________________________________
  bool HitFinder2D::BuildHit(Hit* h)
  {
    const int n0  = fHit.fSignal.size() - 2;
    auto [dx, dy] = fHit.GetDxDy(n0);

    // get position correction
    const double xcorr = fHit.GetXcorr(dx, fHit.GetHitClass()) / fParams.padSizeX;
    std::tie(dx, dy)   = CorrectPosition(dx, dy, xcorr);

    // get anode wire offset
    for (int ia = 0; ia <= NANODE; ia++) {
      const float ya = -1.35 + ia * 0.3;  //anode position in local pad coordinates
      if (dy > ya + 0.15) continue;
      break;
    }

    // Error parametrization X : parabolic model on cl size
    const double parX[] = {0.713724, -0.318667, 0.0366036};
    const double parY[] = {0.0886413, 0., 0.0435141};
    const int nex       = std::min(n0, 7);
    const double edx    = (n0 < 3) ? 1. : parX[0] + parX[1] * nex + parX[2] * nex * nex;
    const double edy    = (n0 < 3) ? 1. : parY[0] + parY[2] * dy * dy;
    const double edt    = (n0 < 3) ? 60. : 26.33;  // should be parametrized as function of da TODO

    // COMPUTE TIME
    double t_avg = 0.;
    for (int idx = 1; idx <= n0; idx++) {
      HitFactory2D::signal& sig = fHit.fSignal[idx];
      const double dtFEE        = fgDT[0] * (sig.s - fgDT[1]) * (sig.s - fgDT[1]) / fClk;
      t_avg += (sig.t - dtFEE);
      if (sig.xe > 0) sig.x += dy / fParams.padSizeY;
    }

    const double time   = (n0 > 1) ? t_avg / n0 : -21.;
    const double tdrift = 100.;  // should depend on Ua

    // COMPUTE ENERGY (conversion to GeV)
    const double e_avg = 1.e-9
                         * std::accumulate(fHit.fSignal.begin(), fHit.fSignal.end(), 0,
                                           [](int sum, const auto& sig) { return sum + sig.s; });

    CalibrateHit(h, dx, dy, edx, edy, edt, time, tdrift, e_avg);
    return kTRUE;
  }

  //_______________________________________________________________________________
  std::pair<double, double> HitFinder2D::CorrectPosition(double dx, double dy, const double xcorr)
  {
    dx -= xcorr;
    fHit.RecenterXoffset(dx);
    dy = dx - dy;
    fHit.RecenterYoffset(dy);

    const double ycorr = fHit.GetYcorr(dy) / fParams.padSizeY;
    dy += ycorr;
    fHit.RecenterYoffset(dy);
    dx *= fParams.padSizeX;
    dy *= fParams.padSizeY;
    return std::make_pair(dx, dy);
  }

  //_______________________________________________________________________________
  void HitFinder2D::CalibrateHit(Hit* h, const double dx, const double dy, const double edx, const double edy,
                                 const double edt, const double time, const double tdrift, const double eloss)
  {
    const ROOT::Math::XYZVector& local_pad = fParams.rowPar[fHit.vrM].padPar[fHit.vcM].pos;
    const ROOT::Math::XYZVector local(local_pad.X() + dx, local_pad.Y() + dy, local_pad.Z());
    const ROOT::Math::XYZVector global = fParams.translation + fParams.rotation(local);
    h->SetX(global.X());
    h->SetY(global.Y());
    h->SetZ(global.Z());

    h->SetDx(edx);
    h->SetDy(edy);
    h->SetDz(0.);
    h->SetDxy(0.);
    h->SetTime(fClk * (fHit.vt0 + time) - tdrift + 30.29 + fHitTimeOff, edt);
    h->SetELoss(eloss);
    h->SetClassType();
    h->SetMaxType(fHit.IsMaxTilt());
    h->SetOverFlow(fHit.HasOvf());
  }

  //_______________________________________________________________________________
  Hit HitFinder2D::MakeHit(int ic, const Cluster2D* cl, std::vector<const CbmTrdDigi*>* digis, size_t nclusters)
  {
    if (fDigis.empty()) {
      fDigis.resize(nclusters);
    }

    Hit hit;
    if (!LoadDigis(digis, ic)) return hit;
    if (!ProjectDigis(ic)) return hit;

    hit.SetAddress(fParams.address);
    hit.SetRefId(ic);
    BuildHit(&hit);
    return hit;
  }

  //_______________________________________________________________________________
  int HitFinder2D::LoadDigis(vector<const CbmTrdDigi*>* din, int cid)
  {
    /** Load RAW digis info in calibration aware strucuture CbmTrdDigiReco
 * Do basic sanity checks; also incomplete adjacent digi and if found merge them.
 */

    int last_col(-1);
    for (auto i = din->begin(); i != din->end(); i++) {
      if ((*i) == nullptr) continue;
      int colT(-1), colR(-1);
      const CbmTrdDigi* dgT = (*i);
      GetPadRowCol(dgT->GetAddressChannel(), colT);

      // check column order for cluster  /// TO DO: we can probably drop this check
      if (last_col >= 0 && colT != last_col + 1) {
        L_(error) << "TrdModuleRec2D::LoadDigis : digis in cluster " << cid << " not in increasing order !";
        return 0;
      }

      last_col              = colT;
      const CbmTrdDigi* dgR = nullptr;

      auto j = std::next(i);
      if (j != din->end()) {
        dgR = (*j);
        GetPadRowCol(dgR->GetAddressChannel(), colR);
      }
      if (colR == colT && dgR != nullptr) {
        fDigis[cid].emplace_back(new DigiRec(*dgT, *dgR));
        (*j) = nullptr;
      }
      else
        fDigis[cid].emplace_back(new DigiRec(*dgT));
    }
    return fDigis[cid].size();
  }

  //_______________________________________________________________________________
  int HitFinder2D::ProjectDigis(int cid, int cjd)
  {
    /** Load digis information in working vectors Digis are represented in the normal coordinate system of
   * (pad width [pw], DAQ time [clk], signal [ADC chs]) with rectangular signals at integer
   * positions.
   */

    if (fDigis[cid].empty()) {
      L_(debug) << "TrdModuleRec2D::ProjectDigis : Request cl id " << cid << " not found.";
      return 0;
    }

    std::vector<HitFactory2D::signal>& hitSig = fHit.fSignal;
    fHit.reset();

    bool on(0);              // flag signal transmition
    int n0(0), nsr(0),       // no of signals in the cluster (all/rect)
      NR(0),                 // no of rect signals/channel in case of RC
      NT(0),                 // no of tilt signals/channel in case of RC
      ovf(1);                // over-flow flag for at least one of the digis
    Char_t dt0(0);           // cluster time offset wrt arbitrary t0
    double err,              // final error parametrization for signal
      xc(-0.5),              // running signal-pad position
      max(0.);               // max signal
    const double re = 100.;  // rect signal
    const double te = 100.;  // tilt signal

    int j(0), col(-1), col0(0), col1(0), step(0), row1;

    // link second row if needed
    vector<DigiRec*>::iterator i1;
    if (cjd >= 0) {
      if (fDigis[cjd].empty()) {
        L_(debug) << "TrdModuleRec2D::ProjectDigis : Request cl id " << cjd << " not found. Skip second row.";
        cjd = -1;
      }
      else
        i1 = fDigis[cjd].begin();
    }

    for (auto i = fDigis[cid].begin(); i != fDigis[cid].end(); i++, j++) {
      const DigiRec* dg  = (*i);
      const DigiRec* dg0 = nullptr;
      const DigiRec* dg1 = nullptr;

      //  initialize
      if (col < 0) {
        fHit.vrM = GetPadRowCol(dg->GetAddressChannel(), col);
        fHit.vt0 = dg->GetTimeDAQ();  // set arbitrary t0 to avoid double digis loop
      }
      GetPadRowCol(dg->GetAddressChannel(), col0);
      int nt = 0;  // no of tilt signals/channel in case of RC
      int nr = 0;  // no of rect signals/channel in case of RC

      // read calibrated signals
      double t = dg->GetTiltCharge(on);
      if (on) nt = 1;
      double r = dg->GetRectCharge(on);
      if (on) nr = 1;
      // look for matching neighbor digis in case of pad row cross
      if (cjd >= 0) {
        if ((dg0 = (i1 != fDigis[cjd].end()) ? (*i1) : nullptr)) {
          row1 = GetPadRowCol(dg0->GetAddressChannel(), col1);
          if (!step) step = fHit.vrM - row1;
          if (col1 == col0) {
            r += dg0->GetRectCharge(on);
            if (on) nr++;
          }
          else
            dg0 = nullptr;
        }
        if (step == 1 && i1 != fDigis[cjd].begin()) {
          dg1 = (*(i1 - 1));
          GetPadRowCol(dg1->GetAddressChannel(), col1);
          if (col1 == col0 - 1) {
            t += dg1->GetTiltCharge(on);
            if (on) nt++;
          }
          else
            dg1 = nullptr;
        }
        if (step == -1 && i1 != fDigis[cjd].end() && i1 + 1 != fDigis[cjd].end()) {
          dg1 = (*(i1 + 1));
          GetPadRowCol(dg1->GetAddressChannel(), col1);
          if (col1 == col0 + 1) {
            t += dg1->GetTiltCharge(on);
            if (on) nt++;
          }
          else
            dg1 = nullptr;
        }
        if (dg0) i1++;
      }

      //TO DO: two blocks below might be mergable
      // process tilt signal/time
      char ddt = dg->GetTiltTime() - fHit.vt0;  // signal time offset wrt prompt
      if (ddt < dt0) dt0 = ddt;
      if (abs(t) > 0) {
        if (nt > 1) t *= 0.5;
        err = te * (nt > 1 ? 0.707 : 1);
        if (dg->HasTiltOvf()) {
          ovf = -1;
          err = 150.;
        }
        if (t > max) {
          max      = t;
          fHit.vcM = j;
          fHit.SetMaxTilt(1);
          fHit.viM = hitSig.size();
        }
      }
      else
        err = 300.;

      hitSig.emplace_back(t, err, ddt, xc, 0.035);
      xc += 0.5;

      // process rect signal/time
      ddt = dg->GetRectTime() - fHit.vt0;  // signal time offset wrt prompt
      if (ddt < dt0) dt0 = ddt;
      if (abs(r) > 0) {
        nsr++;
        if (nr > 1) r *= 0.5;
        err = re * (nr > 1 ? 0.707 : 1);
        if (dg->HasRectOvf()) {
          ovf = -1;
          err = 150.;
        }
        if (r > max) {
          max      = r;
          fHit.vcM = j;
          fHit.SetMaxTilt(0);
          fHit.viM = hitSig.size();
        }
      }
      else
        err = 300.;

      hitSig.emplace_back(r, err, ddt, xc, 0.);
      xc += 0.5;
      NR += nr;
      NT += nt;
    }

    //TO DO: two blocks below might be mergable
    // add front and back anchor points if needed
    if (std::abs(hitSig[0].s) > 1.e-3) {
      xc       = hitSig[0].x;
      char ddt = hitSig[0].t;
      hitSig.emplace(hitSig.begin(), 0., 300., ddt, xc - 0.5, 0.);
      fHit.viM++;
    }
    int n(hitSig.size() - 1);
    if (std::abs(hitSig[n].s) > 1.e-3) {
      xc       = hitSig[n].x + 0.5;
      char ddt = hitSig[n].t;
      hitSig.emplace_back(0., 300., ddt, xc, 0.035);
    }

    n0 = hitSig.size() - 2;
    // compute cluster asymmetry
    int nR = n0 + 1 - fHit.viM;
    if (nR == fHit.viM) {
      fHit.SetSymmHit(1);
      if (hitSig.size() % 2) {
        double LS(0.), RS(0.);
        for (UChar_t idx(0); idx < fHit.viM; idx++)
          LS += hitSig[idx].s;
        for (uint idx(fHit.viM + 1); idx < hitSig.size(); idx++)
          RS += hitSig[idx].s;
        fHit.SetLeftSgn(LS < RS ? 0 : 1);
      }
    }
    else {
      fHit.SetSymmHit(0);
      if (fHit.viM < nR)
        fHit.SetLeftHit(0);
      else if (fHit.viM > nR)
        fHit.SetLeftHit(1);
    }
    // recenter time and space profile
    fHit.vt0 += dt0;
    for (auto& sig : hitSig) {
      sig.t -= dt0;
      sig.x -= fHit.vcM;
    }
    fHit.vcM += col;

    // check if all signals have same significance
    int nmiss = 2 * nsr - NR;
    if (cjd >= 0 && nmiss) {
      fHit.SetBiasX(1);
      for (UChar_t idx(1); idx < fHit.viM; idx++) {
        if (hitSig[idx].xe > 0.) continue;  //  select rect signals
        if (hitSig[idx].se > re * 0.8) fHit.SetBiasXleft(1);
      }
      if (hitSig[fHit.viM].xe <= 0. && hitSig[fHit.viM].se > re * 0.8) fHit.SetBiasXmid(1);
      for (UChar_t idx(fHit.viM + 1); idx < hitSig.size() - 1; idx++) {
        if (hitSig[idx].xe > 0.) continue;  //  select rect signals
        if (hitSig[idx].se > re * 0.8) fHit.SetBiasXright(1);
      }
    }
    else
      fHit.SetBiasX(0);
    nmiss = 2 * n0 - 2 * nsr - NT;
    if (cjd >= 0 && nmiss) {
      fHit.SetBiasY();
      for (UChar_t idx(1); idx < fHit.viM; idx++) {
        if (hitSig[idx].xe > 0. && hitSig[idx].se > te * 0.8) fHit.SetBiasYleft(1);  //  select tilt signals
      }
      if (hitSig[fHit.viM].xe > 0. && hitSig[fHit.viM].se > te * 0.8) fHit.SetBiasYmid(1);
      for (UChar_t idx(fHit.viM + 1); idx < hitSig.size() - 1; idx++) {
        if (hitSig[idx].xe > 0. && hitSig[idx].se > te * 0.8) fHit.SetBiasYright(1);  //  select tilt signals
      }
    }
    else
      fHit.SetBiasY(0);

    if (ovf < 0) fHit.SetOvf();
    return ovf * (hitSig.size() - 2);
  }

  int HitFinder2D::GetPadRowCol(int address, int& c)
  {
    c = address % fParams.rowPar[0].padPar.size();
    return address / fParams.rowPar[0].padPar.size();
  }

  ///////////////////////////////////////////

  //_______________________________________________________________________________
  std::pair<double, double> HitFactory2D::GetDxDy(const int n0)
  {
    double dx, dy;
    switch (n0) {
      case 1:
        if (IsMaxTilt()) {  // T
          dx = -0.5;
          dy = 0;
        }
        else {  // R
          dx = 0.5;
          dy = 0;
        }
        break;
      case 2:
        if (IsOpenLeft() && IsOpenRight()) {  // RT
          dx = viM == 1 ? 0. : -1;
          dy = -0.5;
        }
        else {  // TR
          dx = 0.;
          dy = 0.5;
        }
        break;
      case 3:
        if (IsMaxTilt() && !IsSymmHit()) {  // TRT asymm
          dx = viM == 1 ? 0. : -1;
          dy = GetYoffset();
        }
        else if (!IsMaxTilt() && IsSymmHit()) {  // TRT symm
          dx = 0.;
          dy = GetYoffset();
        }
        else if (IsMaxTilt() && IsSymmHit()) {  // RTR symm
          dx = GetXoffset();
          dy = 0.;
        }
        else if (!IsMaxTilt() && !IsSymmHit()) {  // RTR asymm
          dx = GetXoffset();
          dy = viM == 1 ? -0.5 : 0.5;
        }
        break;
      default:
        dx = GetXoffset();
        dy = GetYoffset();
        break;
    }
    RecenterXoffset(dx);
    return std::make_pair(dx, dy);
  }

  //_______________________________________________________________________________
  double HitFactory2D::GetXoffset(int n0) const
  {
    double dx(0.), R(0.);
    int n(n0 ? n0 : fSignal.size());
    for (int ir(0); ir < n; ir++) {
      if (fSignal[ir].xe > 0) continue;  // select rectangular coupling
      R += fSignal[ir].s;
      dx += fSignal[ir].s * fSignal[ir].x;
    }
    if (std::abs(R) > 0) return dx / R;
    //L_(debug) << "HitFactory2D::GetXoffset : Null total charge for hit size " << n;
    return 0.;
  }

  //_______________________________________________________________________________
  double HitFactory2D::GetYoffset(int n0) const
  {
    double dy(0.), T(0.);
    int n(n0 ? n0 : fSignal.size());
    for (int it(0); it < n; it++) {
      if (fSignal[it].xe > 0) {  // select tilted coupling
        T += fSignal[it].s;
        dy += fSignal[it].s * fSignal[it].x;
      }
    }
    if (std::abs(T) > 0) return dy / T;
    // L_(debug) << "HitFactory2D::GetYoffset : Null total charge for hit size " << n;
    //if (CWRITE(1))
    return 0.;
  }

  //_______________________________________________________________________________
  void HitFactory2D::RecenterXoffset(double& dx)
  {
    /** Shift graph representation to fit dx[pw] in [-0.5, 0.5]
   */

    if (dx >= -0.5 && dx < 0.5) return;
    int ishift = int(dx - 0.5) + (dx > 0.5 ? 1 : 0);
    if (vcM + ishift < 0)
      ishift = -vcM;
    else if (vcM + ishift >= nCols)
      ishift = nCols - vcM - 1;

    dx -= ishift;
    vcM += ishift;
    for (uint idx(0); idx < fSignal.size(); idx++)
      fSignal[idx].x -= ishift;
  }

  //_______________________________________________________________________________
  void HitFactory2D::RecenterYoffset(double& dy)
  {
    /** Shift graph representation to fit dy[ph] in [-0.5, 0.5]
   */

    if (dy >= -0.5 && dy < 0.5) return;
    int ishift = int(dy - 0.5) + (dy > 0.5 ? 1 : 0);
    dy -= ishift;
  }

  //_______________________________________________________________________________
  int HitFactory2D::GetHitClass() const
  {
    /** Incapsulate hit classification criteria based on signal topology
 * [0] : center hit type
 * [1]  : side hit type
 */

    int n0(fSignal.size() - 2);
    if ((n0 == 5 && ((IsMaxTilt() && IsSymmHit()) || (!IsMaxTilt() && !IsSymmHit()))) ||  // TRTRT symm/asymm
        n0 == 4 || (n0 == 3 && ((IsMaxTilt() && IsSymmHit()) || (!IsMaxTilt() && !IsSymmHit()))))
      return 1;  // RTR symm/asymm
    else if (n0 > 5 && HasOvf())
      return 2;
    return 0;
  }

  //_______________________________________________________________________________
  int HitFactory2D::GetHitRcClass(int a0) const
  {
    int a0m      = std::abs(a0);
    uint8_t xmap = vyM & 0xff;
    if (a0m == 2 && IsBiasXleft() && IsBiasXright() && !IsBiasXmid())
      return 0;
    else if (a0m == 3 && ((IsBiasXleft() && IsBiasXright()) || xmap == 116 || xmap == 149 || xmap == 208))
      return 1;
    else if (!IsBiasXleft()
             && (a0m == 2
                 || (a0m == 3 && ((!IsBiasXright() && IsBiasXmid()) || xmap == 209 || xmap == 212 || xmap == 145))))
      return 2;
    else if (!IsBiasXright()
             && (a0m == 2 || (a0m == 3 && ((!IsBiasXleft() && IsBiasXmid()) || xmap == 112 || xmap == 117))))
      return 3;
    else
      return -1;
  }

  //_______________________________________________________________________________
  double HitFactory2D::GetXcorr(double dxIn, int typ, int cls) const
  {
    /** Give the linear interpolation of SYS correction for current position offset "dx" based
 * on LUT calculated wrt MC EbyE data. The position offset is expresed in [pw] units
 * while the output is in [cm]
 */

    if (typ < 0 || typ > 2) {
      //L_(error)<< GetName() << "::GetXcorr : type in-param "<<typ<<" out of range.";
      return 0;
    }
    double dx = std::abs(dxIn);
    int ii    = std::max(0, Nint(dx / fgCorrXdx) - 1), i0;  //  i1;

    if (ii < 0 || ii > NBINSCORRX) {
      // L_(debug) << GetName() << "::GetXcorr : LUT idx " << ii << " out of range for dx=" << dxIn;
      return 0;
    }
    if (dx < fgCorrXdx * ii) {
      i0 = std::max(0, ii - 1);
      /*i1=ii;*/
    }
    else {
      i0 = ii;
      /*i1=TMath::Min(NBINSCORRX-1,ii+1);*/
    }

    float* xval = &fgCorrXval[typ][i0];
    if (cls == 1)
      xval = &fgCorrRcXval[typ][i0];
    else if (cls == 2)
      xval = &fgCorrRcXbiasXval[typ][i0];
    double DDx = (xval[1] - xval[0]), a = DDx / fgCorrXdx, b = xval[0] - DDx * (i0 + 0.5);
    return (dxIn > 0 ? 1 : -1) * b + a * dxIn;
  }

  //_______________________________________________________________________________
  double HitFactory2D::GetYcorr(double dy, int /* cls*/) const
  {
    /** Process y offset. Apply systematic correction for y (MC derived).
 * The position offset is expresed in [pw] units while the output is in [cm]
 */
    float fdy(1.), yoff(0.);
    int n0(fSignal.size() - 2);
    switch (n0) {
      case 3:
        fdy  = fgCorrYval[0][0];
        yoff = fgCorrYval[0][1];
        if (IsMaxTilt() && IsSymmHit()) {
          fdy  = 0.;
          yoff = (dy > 0 ? -1 : 1) * 1.56;
        }
        else if (!IsMaxTilt() && !IsSymmHit()) {
          fdy  = 0.;
          yoff = (dy > 0 ? -1 : 1) * 1.06;
        }
        else if (!IsMaxTilt() && IsSymmHit()) {
          fdy  = 2.114532;
          yoff = -0.263;
        }
        else /*if(IsMaxTilt()&&!IsSymmHit())*/ {
          fdy  = 2.8016010;
          yoff = -1.38391;
        }
        break;
      case 4:
        fdy  = fgCorrYval[1][0];
        yoff = fgCorrYval[1][1];
        if ((!IsMaxTilt() && IsLeftHit()) || (IsMaxTilt() && !IsLeftHit())) yoff *= -1;
        break;
      case 5:
      case 7:
      case 9:
      case 11:
        fdy  = fgCorrYval[2][0];
        yoff = fgCorrYval[2][1];
        break;
      case 6:
      case 8:
      case 10:
        fdy  = fgCorrYval[3][0];
        yoff = fgCorrYval[3][1];
        if ((!IsMaxTilt() && IsLeftHit()) || (IsMaxTilt() && !IsLeftHit())) yoff *= -1;
        break;
    }
    return dy * fdy + yoff;
  }

  //_______________________________________________________________________________
  bool HitFactory2D::IsOpenRight() const
  {
    int nR = fSignal.size() - 1 - viM;
    return (nR % 2 && IsMaxTilt()) || (!(nR % 2) && !IsMaxTilt());
  }

  float HitFactory2D::fgCorrXdx                 = 0.01;
  float HitFactory2D::fgCorrXval[3][NBINSCORRX] = {
    {-0.001, -0.001, -0.002, -0.002, -0.003, -0.003, -0.003, -0.004, -0.004, -0.006, -0.006, -0.006, -0.007,
     -0.007, -0.008, -0.008, -0.008, -0.009, -0.009, -0.011, -0.011, -0.011, -0.012, -0.012, -0.012, -0.012,
     -0.013, -0.013, -0.013, -0.013, -0.014, -0.014, -0.014, -0.014, -0.014, -0.016, -0.016, -0.016, -0.016,
     -0.017, -0.017, -0.017, -0.018, -0.018, -0.018, -0.018, -0.018, 0.000,  0.000,  0.000},
    {0.467, 0.430, 0.396, 0.364, 0.335, 0.312, 0.291, 0.256, 0.234, 0.219, 0.207, 0.191, 0.172,
     0.154, 0.147, 0.134, 0.123, 0.119, 0.109, 0.122, 0.113, 0.104, 0.093, 0.087, 0.079, 0.073,
     0.067, 0.063, 0.058, 0.053, 0.049, 0.046, 0.042, 0.038, 0.036, 0.032, 0.029, 0.027, 0.024,
     0.022, 0.019, 0.017, 0.014, 0.013, 0.011, 0.009, 0.007, 0.004, 0.003, 0.001},
    {0.001,  0.001,  0.001,  0.001,  0.002,  0.002,  0.001,  0.002,  0.004,  0.003,  0.002,  0.002,  0.002,
     0.002,  0.002,  0.002,  0.003,  0.004,  0.003,  0.004,  0.004,  0.007,  0.003,  0.004,  0.002,  0.002,
     -0.011, -0.011, -0.012, -0.012, -0.012, -0.013, -0.013, -0.013, -0.014, -0.014, -0.014, -0.016, -0.016,
     -0.016, -0.017, -0.017, -0.017, -0.018, -0.018, -0.018, -0.019, 0.029,  0.018,  0.001}};
  float HitFactory2D::fgCorrYval[NBINSCORRY][2]   = {{2.421729, 0.},
                                                   {0.629389, -0.215285},
                                                   {0.23958, 0.},
                                                   {0.151913, 0.054404}};
  float HitFactory2D::fgCorrRcXval[2][NBINSCORRX] = {
    {-0.00050, -0.00050, -0.00150, -0.00250, -0.00250, -0.00350, -0.00450, -0.00450, -0.00550, -0.00650,
     -0.00650, -0.00750, -0.00850, -0.00850, -0.00850, -0.00950, -0.00950, -0.00950, -0.01050, -0.01150,
     -0.01150, -0.01150, -0.01250, -0.01250, -0.01250, -0.01250, -0.01350, -0.01350, -0.01350, -0.01350,
     -0.01450, -0.01450, -0.01450, -0.01550, -0.01550, -0.01550, -0.01550, -0.01650, -0.01650, -0.01550,
     -0.01650, -0.01614, -0.01620, -0.01624, -0.01626, -0.01627, -0.01626, -0.01624, -0.01620, -0.01615},
    {0.36412, 0.34567, 0.32815, 0.31152, 0.29574, 0.28075, 0.26652, 0.25302, 0.24020, 0.22803,
     0.21647, 0.21400, 0.19400, 0.18520, 0.17582, 0.16600, 0.14600, 0.13800, 0.14280, 0.14200,
     0.13400, 0.12600, 0.12200, 0.11000, 0.10200, 0.09400, 0.09000, 0.08600, 0.08200, 0.07400,
     0.07000, 0.06600, 0.06600, 0.06200, 0.05800, 0.05400, 0.05400, 0.05000, 0.04600, 0.04600,
     0.04200, 0.03800, 0.03800, 0.03400, 0.03400, 0.03000, 0.03000, 0.02600, 0.02200, 0.02200}};
  float HitFactory2D::fgCorrRcXbiasXval[3][NBINSCORRX] = {
    {0.00100, 0.00260, 0.00540, 0.00740, 0.00900, 0.01060, 0.01300, 0.01460, 0.01660, 0.01900,
     0.02060, 0.02260, 0.02420, 0.02700, 0.02860, 0.02980, 0.03220, 0.03340, 0.03540, 0.03620,
     0.03820, 0.04020, 0.04180, 0.04340, 0.04460, 0.04620, 0.04740, 0.04941, 0.05088, 0.05233,
     0.05375, 0.05515, 0.05653, 0.05788, 0.05921, 0.06052, 0.06180, 0.06306, 0.06430, 0.06551,
     0.06670, 0.06786, 0.06901, 0.07012, 0.07122, 0.07229, 0.07334, 0.07436, 0.07536, 0.07634},
    {0.00100, 0.00380, 0.00780, 0.00900, 0.01220, 0.01460, 0.01860, 0.01940, 0.02260, 0.02540,
     0.02820, 0.03060, 0.03220, 0.03660, 0.03980, 0.04094, 0.04420, 0.04620, 0.04824, 0.04980,
     0.05298, 0.05532, 0.05740, 0.05991, 0.06217, 0.06500, 0.06540, 0.06900, 0.07096, 0.07310,
     0.07380, 0.07729, 0.07935, 0.08139, 0.08340, 0.08538, 0.08734, 0.08928, 0.08900, 0.09307,
     0.09493, 0.09340, 0.09858, 0.09620, 0.09740, 0.10386, 0.09980, 0.10726, 0.10892, 0.11056},
    {0.00011, 0.00140, 0.00340, 0.00420, 0.00500, 0.00620, 0.00820, 0.00860, 0.01060, 0.01100,
     0.01220, 0.01340, 0.01500, 0.01540, 0.01700, 0.01820, 0.01900, 0.02060, 0.02180, 0.02260,
     0.02340, 0.02420, 0.02500, 0.02500, 0.02660, 0.02740, 0.02820, 0.02900, 0.03020, 0.03180,
     0.03300, 0.03260, 0.03380, 0.03460, 0.03500, 0.03580, 0.03780, 0.03820, 0.03860, 0.03900,
     0.04100, 0.04180, 0.04060, 0.04300, 0.04340, 0.04340, 0.04380, 0.04460, 0.04580, 0.04540}};

}  // namespace cbm::algo::trd
